use base64::{decode, encode};
use std::error::Error;
use std::io;
use std::net::{IpAddr, SocketAddr};
use std::time::Duration;
use std::time::Instant;

#[derive(Debug, PartialEq, Eq, Hash, Copy, Clone)]
pub enum ECCKey {
    PublicKey([u8; 32]),
    PrivateKey([u8; 32]),
}

pub trait Base64Backed {
    fn from_bytes(bytes: [u8; 32]) -> Self;
    fn bytes(&self) -> &[u8; 32];
    fn from_base64<I: AsRef<str>>(key: I) -> io::Result<Self>
    where
        Self: Sized,
    {
        let key = match decode(key.as_ref()) {
            Ok(key) => key,
            _ => {
                return Err(io::Error::new(
                    io::ErrorKind::InvalidData,
                    "Failed to decode base64",
                ))
            }
        }; /*.map_err(|err| {

           })?;*/
        let mut bytes = [0; 32];
        if key.len() != 32 {
            return Err(io::Error::new(io::ErrorKind::Other, "Mismatched key size"));
        }
        bytes.copy_from_slice(&key);
        Ok(Self::from_bytes(bytes))
    }

    fn as_base64(&self) -> io::Result<String> {
        Ok(encode(self.bytes()))
    }
}

impl Base64Backed for ECCKey {
    fn bytes(&self) -> &[u8; 32] {
        match self {
            ECCKey::PublicKey(bytes) => &bytes,
            ECCKey::PrivateKey(bytes) => &bytes,
        }
    }

    fn from_bytes(bytes: [u8; 32]) -> ECCKey {
        ECCKey::PublicKey(bytes)
    }
}

impl ECCKey {
    pub fn extract_public_key(&self) -> Option<ECCKey> {
        //TODO: Determine whether Self is a private key and only the return public part
        Some(self.clone())
    }
}

#[derive(Debug, PartialEq, Eq, Hash, Copy, Clone)]
pub struct SharedKey([u8; 32]);

impl Base64Backed for SharedKey {
    fn bytes(&self) -> &[u8; 32] {
        &self.0
    }
    fn from_bytes(bytes: [u8; 32]) -> SharedKey {
        SharedKey(bytes)
    }
}

#[derive(Debug, Builder, PartialEq, Eq, Hash, Clone)]
pub struct Interface {
    key: ECCKey,
    port: usize,
    fwmark: Option<String>
}

#[derive(Debug, Builder, PartialEq, Eq, Hash, Clone)]
pub struct Peer {
    key: ECCKey,
    #[builder(default = "None")]
    shared_key: Option<SharedKey>,
    #[builder(default = "None")]
    endpoint: Option<SocketAddr>,
    allowed_ips: Vec<(IpAddr, u8)>,
    last_handshake: Option<Duration>,
    #[builder(default = "None")]
    persistent_keepalive: Option<Duration>,
    #[builder(default = "(0u64,0u64)")]
    traffic: (u64, u64),
    #[builder(default = "Instant::now()")]
    parsed: Instant,
}

impl PeerBuilder {
    fn validate(&self) -> Result<(), String> {
        if let Some(ref key) = self.key {
            Ok(())
        } else {
            Err("No key supplied".into())
        }
    }

    pub fn is_whole(&self) -> bool {
        self.validate().is_ok()
    }

    pub fn add_allowed_ip(&mut self, ip: (IpAddr, u8)) {
        if let Some(ref mut ips) = &mut self.allowed_ips {
            ips.push(ip);
        } else {
            self.allowed_ips = Some(vec![ip]);
        }
    }

    pub fn add_last_handshake(&mut self, d: Duration) {
        if !self.last_handshake.is_some() {
            self.last_handshake = Some(Some(d));
        } else {
            self.last_handshake = self
                .last_handshake
                .map(|shake| shake.map(|shake| shake + d));
        }
    }
}

pub trait WireguardController {
    fn peers<'a>(&'a mut self) -> io::Result<Box<Iterator<Item = io::Result<Peer>> + 'a>>;

    fn interface(&mut self) -> io::Result<Interface>;

    fn update_peer(&mut self, peer: &Peer) -> io::Result<()>;
}
